/*********************************************************/
/*  This source code is copyrighted by the author        */
/*  Please read COPYING  for the GNU licence policy.     */
/*********************************************************/
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "tetX.h"
#include "egafont.h"

#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif

#ifndef P_tmpdir
#define P_tmpdir "/tmp"
#endif

extern Display  *display;
extern Visual   *visual;
extern GC	graphContext;
extern Pixmap	screenContext;

extern int	ScreenDepth;
extern Bool	DoInvalidate;
extern unsigned long PixValues[];

static char *FontData;
static int  FontHeight;

int	wide_chars;


int egaShowText(char *text, int x, int y, int fgcol, int bgcol)
{  
   XImage	*xim = NULL;    
   unsigned char *data = NULL;
   unsigned char mask;
   unsigned int  txtwidth;
   int		x1, x2, y1, len;	
   unsigned char fbyte;
   unsigned char *fontptr;
   unsigned long fgrPix, bgrPix;
   tchar    *textbuf, textchar;     
   int	    rc = 4;

#ifdef TET_USE_WIDE_CHARS
   len = mbstowcs(NULL, text, 0);
   textbuf = malloc((len+1) * sizeof(wchar_t));
   if (textbuf == NULL) goto Finito;
   mbstowcs(textbuf, text, len+1);
#else   
   textbuf = text;
   len  = strlen(text);
#endif

   fgrPix = PixValues[fgcol];
   bgrPix = PixValues[bgcol];  

   txtwidth = len * 8;    
   xim = XCreateImage(display, visual, ScreenDepth,
                      ZPixmap, 0, NULL, 
                      txtwidth, FontHeight, 8, 0);
   if (!xim) goto Finito;
   
   data = malloc(xim->bytes_per_line * FontHeight);
   if (!data) goto Finito;
   xim->data = data;

   x1 = 0;
   while ((textchar = *text++) != '\0')
   {  
      fontptr = FontData + textchar*FontHeight;

	  for (y1=0; y1<FontHeight; y1++)      
      {  fbyte = *fontptr++;
		 mask = 0x80;              
         x2 = x1;
		 while (mask)
		 {  XPutPixel (xim, x2, y1, (fbyte & mask) ? fgrPix: bgrPix);
		   	x2++;
           	mask = mask >> 1;
		 }
	  }	   
  
	  x1+=8;

   }      

   XPutImage(display, screenContext, graphContext, xim,
          0, 0, x, y, txtwidth, FontHeight);

   if (DoInvalidate)
   {	RECT rect;
		rect.left = x; rect.right = x+txtwidth;
    	rect.top = y;  rect.bottom = y+FontHeight;
		ContextRectToScreen(&rect);
   }     	  

   rc = 0;

Finito:   
   if (xim) XDestroyImage(xim);   
#ifdef TET_USE_WIDE_CHARS
   if (textbuf) free(textbuf); 
#endif
   return rc;

}

int egaShowScaledText(char *text, int x, int y,
                      int fgcol, int bgcol, int hscale, int vscale)
{

   XImage	 *xim = NULL;    
   unsigned char *data = NULL;
   unsigned char mask;
   signed   int  x1, y1, x2, y2, x3, y3, cx, cy, len;
   unsigned int  textwid, scalewid, scalehei;
   unsigned char fbyte;
   unsigned char *fontptr;
   tchar    *textbuf, textchar;     
   unsigned long fgrPix, bgrPix, curPix;
   int	    rc = 4;

#ifdef TET_USE_WIDE_CHARS
   len = mbstowcs(NULL, text, 0);
   textbuf = malloc((len+1) * sizeof(wchar_t));
   if (textbuf == NULL) goto Finito;
   mbstowcs(textbuf, text, len+1);
#else   
   textbuf = text;
   len  = strlen(text);
#endif

   fgrPix = PixValues[fgcol];
   bgrPix = PixValues[bgcol];  

   textwid  = len * 8;    
   scalewid = textwid  * hscale;
   scalehei = FontHeight * vscale;

   xim  = XCreateImage(display, visual, ScreenDepth,
                       ZPixmap, 0, NULL, 
                       scalewid, scalehei, 8, 0);
   if (!xim) goto Finito;
   
   data = malloc(xim->bytes_per_line * scalehei);
   if (!data) goto Finito;
   xim->data = data;

   x1 = 0; x2 = 0; x3 = 0;
   while ((textchar = *text++) != '\0')
   {  
      fontptr = FontData + textchar*FontHeight;
      
      y1 = 0;
      y2 = FontHeight;
	  
      while (--y2>=0)      
      {  fbyte = *fontptr++;
		 mask = 0x80;              
         x2 = x1;
		 while (mask)
		 {  y3 = y1;
	    	curPix = (fbyte & mask) ? fgrPix: bgrPix;
		    cy = vscale;
		    while(--cy >= 0)
		    {  cx = hscale;
		       x3 = x2;
	    	   while (--cx>=0)
		       {  XPutPixel (xim, x3, y3, curPix);
		          x3++;
	    	   }
		       y3++;
		    }   	  
	    	x2 = x3;
            mask = mask >> 1;
		 }
	 
		 y1 = y3;
      }	   

      x1 = x2;
   }      

   XPutImage(display, screenContext, graphContext, xim,
          0, 0, x, y, scalewid, scalehei);

   if (DoInvalidate)
   {  RECT rect;
      rect.left = x; rect.right = x+scalewid;
      rect.top = y;  rect.bottom = y+scalehei;
      ContextRectToScreen(&rect);
   }     	  
	rc = 0;

Finito:   
   if (xim)  XDestroyImage(xim);   
#ifdef TET_USE_WIDE_CHARS
   if (textbuf) free(textbuf); 
#endif
   return rc;
}


int egaSetFont(char *cFont)
{   int *iFont;
    if (!cFont) return 4;

    iFont = (int *)cFont;
    FontHeight = *iFont;
    
    FontData = cFont + sizeof(int);
    return 0;
}
  
void egaFreeFont(char *cFont)
{
   if (!cFont) return;

   if (cFont + sizeof(int) == FontData)
   {  FontData = NULL; FontHeight = 0; }

   free(cFont);
}        


int egaGetFontHeight(void)
{   return FontHeight; }

int egaGetTextWidth(char *text)
{
   int len;	
#ifdef TET_USE_WIDE_CHARS
    len = mbstowcs(NULL, text, 0);
#else   
    len  = strlen(text);
#endif
	return len * 8;
}

FONT_TYPE egaFontType(void)
{
   return FONT_TYPE_X11_CONSOLE;
}
/*
char *egaGetFontPtr(int chr)
{  return FontData + (chr & 0xff) *FontHeight; }
*/
char *egaLoadFont(char *fontFileName, char *fontDirName, char *mes)
{
   int fd = -1;
   int rc = 4; 
   size_t s;
   off_t  hdrlen;    // File header length

   char  *cFont = NULL;
   int   *iFont;
   char *FontPath, *FontPath1, *tmp_file_name= NULL;
   int   fontheight;

   int   unzipped;
   static const char  *tmp_file_pref = "itet";
   unsigned char strt[4];
   static const char *extensions[] = {"", ".gz", ".psf", ".psf.gz", NULL};	
   const char *extension;	


#ifdef HAVE_DIRENT_H
   
   if (memcmp(fontDirName, "/usr", 4) == 0)
   {
        DIR *DirTest = opendir(fontDirName);

        if (DirTest == NULL)
	{
		DirTest = opendir(fontDirName+4);
         	if (DirTest) strcpy(fontDirName, fontDirName+4);
 	}
   	
	if (DirTest == NULL) return NULL;
	closedir(DirTest);
   }
#endif

   wide_chars = 0;
   FontPath = malloc(strlen(fontDirName) + strlen(fontFileName) + 5);
   strcpy(FontPath, fontDirName);
   fontheight = strlen(FontPath);
   if (FontPath[fontheight-1] != '/') strcat(FontPath, "/");
   strcat(FontPath, fontFileName);
   FontPath1 = GetFullFileName(FontPath);    
   free(FontPath);

   FontPath = malloc(strlen(FontPath1) + 11);
   fontheight = strlen(FontPath1);
   memcpy(FontPath, FontPath1, fontheight); 
   free(FontPath1);
		
   for (unzipped=0; (extension=extensions[unzipped])!=NULL; unzipped++)
   { 		
	strcpy(FontPath + fontheight, extension);
	fd = open(FontPath, O_RDONLY);
	if (fd >= 0) break;
   }

   if (fd < 0)
   {  FontPath[fontheight] = '\0';
      sprintf(mes, "Could not open font %s\n", FontPath);
      free(FontPath);
	  return NULL;
   }		

   fontheight += strlen(extension);
   unzipped = strcasecmp(FontPath+fontheight-3, ".gz");

   if (!unzipped)
   {
	const char  *tmp_dir_name;

	close(fd);  fd = -1;
	tmp_dir_name = getenv("TMPDIR");
	if (tmp_dir_name == NULL) tmp_dir_name = P_tmpdir;

	FontPath1 = malloc(strlen(FontPath) + strlen(tmp_dir_name)
                        + strlen(tmp_file_pref) + 50);
	if (FontPath1 == NULL) goto Ret;

 	 sprintf(FontPath1, "gunzip -c %s > ", FontPath);
	 tmp_file_name = FontPath1 + strlen(FontPath1);
	 sprintf(tmp_file_name, "%s/%s%08x.cfont",
		tmp_dir_name, tmp_file_pref, getpid());
         system(FontPath1);

         fd = open(tmp_file_name, O_RDONLY);
   }       

   if (fd < 0) sprintf(mes, "Could not open font %s\n", FontPath);
   free(FontPath);
   
   if (fd < 0) return NULL;

   /* Try to find the size of characters */
   read(fd, strt, 4);

	/* Removed check for bytes pers char,
       because it will work with wide chars as well */ 
   if (strt[0] == 0x36 && strt[1] == 4) /* && strt[2] <= 1) */
   {  /* psf file */
       fontheight = strt[3];
#ifdef TET_USE_WIDE_CHARS 
	   s = lseek(fd, 0, SEEK_END) - 4;
       lseek(fd, 4, SEEK_SET); 
	   wide_chars = 1;
#else
       s = fontheight<<8;
#endif	   
   }
   else
   {   /* Size of the font file */
       s = (size_t) lseek(fd, 0, SEEK_END);

       fontheight = s>>8;       // Assume 256 characters, FontHeight bytes each
       hdrlen  = s & 255;
       if (hdrlen != 0 && hdrlen != 4) goto Ret;
       s -= hdrlen;

       lseek(fd, hdrlen, SEEK_SET);  // Bypassing header
   }

   cFont = (char *) malloc(s + sizeof(int));
   if (!cFont) goto Ret;

   iFont = (int *)cFont;
   *iFont = fontheight;
   if (read(fd, cFont+sizeof(int), s) == s) rc = 0; 
 
Ret:  
   
   close(fd);
   if (!unzipped)
   {	remove(tmp_file_name);
	free(FontPath1);
   }	

   if (rc)
   {  if (cFont) free(cFont);
      return NULL;
   }

   return cFont;
}


int egaShowCharSpecial(tchar b, int x, int y, int fgcol, int bgcol,
                       int hscale, int vscale, int flags)
{
    XImage	 *xim = NULL;    
    unsigned char *data = NULL;
    unsigned char mask;
   

    signed   int  y1, x2, y2, x3, y3, cx, cy;
    signed   int  xstart, xincr, ystart, yincr;

    unsigned int  scalewid, scalehei;
    unsigned char fbyte;
    unsigned char *fontptr;
    unsigned long fgrPix, bgrPix, curPix;


    fgrPix = PixValues[fgcol];
    bgrPix = PixValues[bgcol];  

    scalewid = hscale * 8;
    scalehei = FontHeight * vscale;

    xim  = XCreateImage(display, visual, ScreenDepth,
                       ZPixmap, 0, NULL, 
                       scalewid, scalehei, 8, 0);
    if (!xim) goto Finito;
   
    data = malloc(xim->bytes_per_line * scalehei);
    if (!data) goto Finito;
    xim->data = data;
    
    if (flags & EGAFONT_HFLIP)
    {   xstart =  FontHeight - 1;
        xincr = -1;
    }
    else
    {  xstart = 0;
       xincr = 1;
    }   

    if (flags & EGAFONT_VFLIP)
    {   ystart =  FontHeight - 1;
        yincr = -1;
    }
    else
    {  ystart = 0;
       yincr = 1;
    }   


    b &= 0xff;
    fontptr = FontData + b*FontHeight;
    y1 = ystart*vscale;
    yincr *= vscale;
    xstart *= hscale;
    xincr *= hscale;

    y2 = FontHeight;
    while (--y2>=0)      
    {    fbyte = *fontptr++;
	 mask = 0x80;              
         x2 = xstart;
	 while (mask)
	 {  y3 = y1;
	    curPix = (fbyte & mask) ? fgrPix: bgrPix;
	    cy = vscale;
	    while(--cy >= 0)
	    {  cx = hscale;
	       x3 = x2;
	       while (--cx>=0)
	       {  XPutPixel (xim, x3, y3, curPix);
	          x3++;
	       }
	       y3++;
	    }   	  
	    x2 += xincr;
            mask = mask >> 1;
	 }
	 
	 y1 += yincr;
     }	   
    


   XPutImage(display, screenContext, graphContext, xim,
          0, 0, x, y, scalewid, scalehei);

   if (DoInvalidate)
   {  RECT rect;
      rect.left = x; rect.right = x+scalewid;
      rect.top = y;  rect.bottom = y+scalehei;
      ContextRectToScreen(&rect);
   }     	  

Finito:   
   if (xim)  XDestroyImage(xim);   
   return 0;
     
} 

XImage *DrawCharOnXImage(char textchar, unsigned long fgrPix, unsigned long bgrPix)
{  

   XImage *xim;
   unsigned char *data = NULL;
   unsigned char mask;
   int		 x2, y1;
   unsigned char fbyte;
   unsigned char *fontptr;

   if (!FontData) return NULL;


   xim  = XCreateImage(display, visual, ScreenDepth,
                       ZPixmap, 0, NULL, 
                       8, FontHeight, 8, 0);
   if (!xim) return NULL;
      
   data = malloc(xim->bytes_per_line * FontHeight);
   if (!data) 
   { XDestroyImage(xim); return NULL ; }
   
   xim->data = data;

   fontptr = FontData + textchar*FontHeight;

   for (y1=0; y1<FontHeight; y1++)       
   {  fbyte = *fontptr++;
      mask = 0x80;              
      x2 = 0;
      while (mask)
      {  XPutPixel (xim, x2, y1, (fbyte & mask) ? fgrPix: bgrPix);
          x2++;
          mask = mask >> 1;
      }
   }

   return xim;
}
