#include <string.h>
#include <stdlib.h>
#include <vga.h>
#include "tetX.h"
#include "tet_values.h"

extern IMAGE_HEADER owl256_Hdr;
extern int ScreenWidth, ScreenHeight, ScreenDepth;
extern unsigned char *ScreenContext;
extern int VideoChipset;


static int StandardEGAPalette[48]=
   {  0,  0,  0,    0,  0, 42,    0, 42,  0,    0, 42, 42,
     42,  0,  0,   42,  0, 42,   42, 21,  0,   42, 42, 42,
     21, 21, 21,   21, 21, 63,   21, 63, 21,   21, 63, 63,
     63, 21, 21,   63, 21, 63,   63, 63, 21,   63, 63, 63
   };

static int *ScreenPal = NULL;

int initPalette(void)
{

   if (ScreenDepth <= 4)
        ScreenPal = StandardEGAPalette;
   else
   {  int  sizeEntry = sizeof(int) * 3;

      ScreenPal = (int *)malloc(sizeEntry * (1 << ScreenDepth));
      if (!ScreenPal) return 8;
      memcpy(ScreenPal, StandardEGAPalette, sizeEntry * 16);

//      vga_getpalvec(0, 16, ScreenPal);
   }

   return 0;
}

void deinitPalette(void)
{
   if (ScreenDepth > 4 && ScreenPal) free(ScreenPal);
}

void   setPalette(IMAGE_HEADER *imHdr, int maxcols)
{
   int used = imHdr->usedcol;
   int PalColsUsed = 16;
   PAL_ENTRY curImPal;
   PAL_ENTRY *ImPal = imHdr->pal;
   unsigned char *PalEquiv = imHdr->PalEqu;
   long dist, sdist;
   short dr, dg, db;
   short i, j, sind;

#ifndef SHOW_256C_IMAGE
   if (ScreenDepth <= 4 && imHdr->imbpp > 4) return;
#endif


   for (i=0; i<used; i++)
   {
      curImPal = ImPal[i];
      sdist = MAXLONG;
      sind = -1;
      for (j=0; j<PalColsUsed; j++)
      {
         dr = ScreenPal[j*3] - curImPal.r;
         dg = ScreenPal[j*3+1] - curImPal.g;
         db = ScreenPal[j*3+2] - curImPal.b;
         dist = dr*dr + dg*dg + db*db;
         if (dist < sdist)
         { sind = j; sdist = dist;
           if (sdist == 0) break;
         }
      }

      if (sind >=0 && (sdist == 0 || PalColsUsed >= maxcols))
          PalEquiv[i] = sind;
      else
      {  // These will never work in 16c mode
         PalEquiv[i] = PalColsUsed;
         j = PalColsUsed; PalColsUsed++;
         ScreenPal[j*3] = curImPal.r;
         ScreenPal[j*3+1] = curImPal.g;
         ScreenPal[j*3+2] = curImPal.b;
      }
    }


    if (ScreenDepth > 4)
         vga_setpalvec(1, PalColsUsed-1, ScreenPal+3);

}



int   showSqueezedImage(IMAGE_HEADER *imHdr, int x, int y, int transp,
                                               int stepx, int stepy)
{
   short wid = imHdr->imwid;
   short hei = imHdr->imhei;
   short bpp = imHdr->imbpp;
   unsigned char *imstart = imHdr->image, *imptr;
   unsigned char *bufptr, *bufline, c, col;
   unsigned char smallbuf[20];
   unsigned char *PalEquiv = imHdr->PalEqu;
   short i, j, k, towid;

#ifndef SHOW_256C_IMAGE
   if (ScreenDepth <= 4 && bpp > 4) return 0;
#endif

   if (ScreenContext)
       bufline = ScreenContext + (ScreenWidth * y) + x;
   else
   if(wid <= 20)
       bufline = smallbuf;
   else
   {   bufline = (unsigned char *) malloc(wid);
       if (!bufline) return 4;
   }


   towid = wid;
   if (bpp <= 4) towid >>= 1;



   for (i=0; i<hei; i+=stepy)
   {   if (!ScreenContext)
              vga_getscansegment(bufline, x, y+i, wid);

       bufptr = bufline;
       imptr = imstart + i*towid;
       for (j=0; j<towid; j+=stepx)
       {  c = '\0';
          for (k=0; k<stepx; k++)  c |= *imptr++;
          if (bpp > 4) col=c;
          else
          {  col = c >> 4;
             if (col != transp)
                     *bufptr = PalEquiv[col];
             bufptr++;
             col = c & 0xf;
          }

          if (col != transp)
                  *bufptr = PalEquiv[col];
          bufptr++;
       }

       if (ScreenContext)
           bufline += ScreenWidth;
       else
           vga_drawscansegment(bufline, x, y+i, wid);
  }

   if (!ScreenContext && (wid > 20)) free(bufline);
   return 0;
}

int   showImage(IMAGE_HEADER *imHdr, int x, int y, int transp)
{
  return showSqueezedImage(imHdr, x, y, transp, 1, 1);
}

